from ._query import (
    sql, Database
)

from . import _pandas # triggers pandas monkeypatching

try: # dumb hack to get this to work on 3.7 today
    from . import _data as data
except:
    pass
    
from ._io import save, load

try: # dumb hack to get this to work on 3.7 today
    import importlib.metadata
    __version__ = importlib.metadata.version(__package__ or __name__)
except:
    __version__ = '0.4.0'



from pandas.core.base import PandasObject

from . import _query
from ._util import _get_if_file

def _insist_single_row(df):
    if len(df) != 1:
        raise ValueError(f'DataFrame has {len(df)} rows, but should only have 1.')

def _insist_single_col(df):
    if len(df.columns) != 1:
        raise ValueError(f'DataFrame has {len(df.columns)} columns, but should only have 1.')

def _insist_single_row_or_col(df):
    if len(df.columns) == 1 or len(df) == 1:
        pass
    else:
        raise ValueError(f'DataFrame should have a single row or column, but has shape f{df.shape}')


def _sql(df, s='', tbl_name='_df'):
    """ Run a DuckDB SQL query against Pandas DataFrame.

    DataFrame will be referred to by string given in `tbl_name`.
    The query starts with a `from {table_name}` clause.
    Since DuckDB 0.6, the `select` clause is optional, defaulting to `select *`.

    Examples
    --------

    ```
    df.sql('select col_a where col_b > 0')
    ```

    is equivalent to the query

    ```
    select
        col_a
    from
        _df
    where
        col_b > 0
    ```


    `df.sql('where col_b > 0')` is equivalent to

    ```
    select
        *
    from
        _df
    where
        col_b > 0
    ```

    Note that `df.sql()` is equivalent to

    ```
    select * from _df
    ```
    """

    s = _get_if_file(s)
    if not _is_CTE(s):
        s = f'from {tbl_name}\n' + s

    tables = {tbl_name: df}

    return _query.sql(s, **tables)

def _is_CTE(s):
    return s.strip().startswith('with')



def _as_list(df):
    """Transform a df with one row or one column to a list"""
    if len(df.columns) == 1:
        col = df.columns[0]
        out = list(df[col])
    elif len(df) == 1:
        out = list(df.loc[0])
    else:
        raise ValueError(f'DataFrame should have a single row or column, but has shape f{df.shape}')

    return out


def _as_dict(df):
    """Transform a df with one row to a dict
    """
    _insist_single_row(df)

    out = dict(df.iloc[0])

    return out

def _as_item(df):
    """Transform a df with one row and one column to single element"""
    _insist_single_row(df)
    _insist_single_col(df)

    out = _as_list(df)[0]

    return out


def _cols2dict(df):
    cols = df.columns

    if len(cols) != 2:
        raise ValueError(f'DataFrame should have 2 columns, but has {len(df.columns)}.')

    k = cols[0]
    v = cols[1]

    k = df[k]
    v = df[v]

    out = dict(zip(k, v))

    return out


PandasObject.sql = _sql

PandasObject.aslist = _as_list
PandasObject.asdict = _as_dict
PandasObject.asitem = _as_item
PandasObject.cols2dict = _cols2dict



from pathlib import Path


def _is_file(s: str) -> bool:
    try:
        return Path(s).is_file()
    except OSError:
        # catches `OSError: [Errno 63] File name too long`
        return False

def _get_if_file(s: str) -> str:
    if _is_file(s):
        with open(s) as f:
            s = f.read()

    return s



import duckdb
import os

def save(filename, **dfs):
    try:
        os.remove(filename)
    except OSError:
        pass

    config = {'enable_external_access': False}
    con = duckdb.connect(database=filename, config=config)

    for tbl_name, df in dfs.items():
        temp_df_name = '__tmp_duckdb__' + tbl_name
        con.register(temp_df_name, df)
        con.execute(f'create table {tbl_name} as select * from {temp_df_name}')

    con.close()


def load(filename):
    config = {'enable_external_access': False}
    con = duckdb.connect(database=filename, config=config, read_only=True)

    tbl_names = list(con.execute('show tables;').df()['name'])

    dfs = {
        tbl_name: con.execute(f'select * from {tbl_name}').df()
        for tbl_name in tbl_names
    }

    con.close()

    return dfs



import duckdb
import pyarrow as pa

from ._util import _get_if_file


def sql(s, **dfs):
    db = Database(**dfs)
    out = db.sql(s)

    return out


class Database:
    def __init__(self, **dfs):
        con = duckdb.connect(
            database = ':memory:',
            config = {'enable_external_access': False},
        )

        for tbl_name, df in dfs.items():
            df = pa.Table.from_pandas(df)
            con.register(tbl_name, df)

        self._con = con

    def __repr__(self):
        tables = _yield_table_lines(self)
        tables = [
            f'\n    {t}'
            for t in tables
        ]
        tables = ''.join(tables)
        tables = tables or ' None'

        out = 'pdx2.Database:' + tables

        return out

    def sql(self, s):
        s = _get_if_file(s)
        out = self._con.execute(s).df()
        return out


def _yield_table_lines(db):
    df = db.sql('show all tables')

    for _, tbl in df.iterrows():
        name = tbl['name']
        columns = list(tbl['column_names'])
        n = db.sql(f'select count() from {name}').asitem()

        yield f'{name}: {n} x {columns}'
